#!/usr/bin/env python
'''
CosEN_para.py is analogus to CosEN.py but in a parallel running fashion.
Input:
    Expression matrix: Expression/filtered_lncRNA_z_EXP.csv
    Drug responses data: drug_res/drug_res_merge.csv
Operation: (*lead to one or multi output)
    1. Parameter Optimization*
        Output: optimized parameter for 265 drugs (IC50 and AUC)
    2. Bootstrapping
        Using optimized parameters, compute the predictive scores for each drug (IC50 and AUC)
    3. Get top predictive genes*
        Sort the genes by predictive scores and get top 20 genes.
        Output: predictive score of all genes for each drug (IC50 and AUC)
    4. Final model construction*:
        i. Get coefficients
            Output: Average coefficiencts of every predictors in each model (IC50 and AUC)
        ii. Model performance assessments
            Pearson's correlation coefficients, kendall's tau
'''

import pp
import sys
import numpy as np
import pandas as pd
import scipy.stats as stats
import matplotlib.pyplot as plt
import matplotlib

from pandas import DataFrame as df
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import ElasticNetCV
from sklearn.linear_model import enet_path


__Author__ = 'Yue Wang'
__Date__ = 'JUl 7, 2017'


def bootPred(current_expr, current_res, op_alpha, op_l1_ratio):
    '''
    Bootstrapping and compute the predictive scores.
    Input:
        current_expr, current_res: pandas dataframe, row = gene/pharm, col = expression/response across cell lines
    Operation:
        1. Array reshape
        2. Bootstrap resample
        3. Compute bootstrapping performance
    Output:
        score_bs: pandas dataframe, row = genes, col = predictive scores
    '''
    # array reshape

    array_expr = np.array(current_expr.values.T)
    array_res = np.ravel(current_res.values)
    score_bs = df(index=[i + 1 for i in range(200)], columns=['Pearsonr coef.', 'Pearsonr p'])

    for i in range(200):
        # resample
        resample_i, expr, res = bootstrapResample(array_expr, array_res, n=None)
        test_expr = np.array([array_expr[j] for j in range(len(array_expr)) if j not in resample_i])
        test_res = np.ravel([[array_res[j] for j in range(len(array_res)) if j not in resample_i]])

        # Training
        enet = ElasticNet(alpha=op_alpha, l1_ratio=op_l1_ratio)
        enet.fit(expr, res)

        # assess the performance of this iteration
        pred_res = enet.predict(test_expr)
        coef, p = stats.pearsonr(test_res, pred_res)
        score_bs.at[i, 'Pearsonr coef.'] = coef
        score_bs.at[i, 'Pearsonr p'] = p

    return score_bs


def bootstrapResample(current_expr, current_res, n=None):
    '''
    A bootstrapping procedure to generate resampled datasets.
    current_expr, current_res: array-like matrix
    n: int, optional, length of resampled array, equal to len(data to resample) if n==None
    Results: resampled_expr, resampled_res
    '''
    # assign bootstrap resampled size
    if n is None:
        assert len(current_expr) == len(current_res)
        n = len(current_res)

    # generate resample index
    resample_i = np.floor(np.random.rand(n) * n).astype(int)
    resampled_expr = np.array([current_expr[i] for i in resample_i], dtype=float)
    resampled_res = [current_res[i] for i in resample_i]

    return resample_i, resampled_expr, resampled_res


def paraEN(current_expr, current_res, pharm, drug):
    '''
    Parameter optimization for elastic net regression model.
    ONLY require for models with genes selected by bootstrapping.
    Input:
        gene_expr: pandas dataframe, row = gene, col = gene expression matrix across all cell lines
        drug_res: pandas dataframe, row = pharm, col = drug response matrix across all cell lines
    Return:
        model.alpha_, model.l1_ratio_
    '''
    print 'Optimizing alpha and l1 ratio...'

    model = ElasticNetCV(
        l1_ratio=[.28, .36, .44, .52, .6, .68, .76, .84, .92, 1.0],
        n_alphas=200,
        max_iter=10000000,
        cv=10,
        precompute='auto').fit(
        np.array(current_expr.values.T),
        np.ravel(current_res.values))

    alphas_enet, coef_enet, _ = enet_path(
        np.array(current_expr.values.T),
        np.ravel(current_res.values),
        l1_ratio=model.l1_ratio_)

    # display optimization path
    fig1 = plt.figure()
    ax = plt.gca()
    ax.set_color_cycle(2 * ['b', 'r', 'g', 'c', 'k'])
    plt.plot(-np.log10(alphas_enet), coef_enet.T)
    plt.xlabel('-Log(alpha)')
    plt.ylabel('Coefficients')
    plt.axis('tight')
    fig1.savefig('Elastic_net_para/optimization_path/' + pharm + '/' + str(drug) + '.pdf')

    # out put the parameters
    return model.alpha_, model.l1_ratio_


def runEN(grouped_res, drug_id, expr, sub_pharm, job_id):
    '''
    Skeletal of elastic net regression model: training, bootstrapping, construction
    Input:
        grouped_res: hierarchical pandas dataframe, index = drug, cell line, col = cell lines, value = IC50 or AUC
        drug_id: list, cosmic of drugs in this running
        expr: pandas dataframe, index = genes, col = cell lines, value = expression
        sub_pharm: list, 'LN_IC50' and 'AUC'
        job_id: str, parallel parameter indicating the id of processor used
    Output:
        final_model_performance: csv file, performance of all of the models in this running, given job_id
        predictive_scores: csv file
        final model_coefficients: csv file
    '''
    matplotlib.rcParams['pdf.fonttype'] = 42
    matplotlib.rcParams['ps.fonttype'] = 42
    print 'Submitting to processor' + job_id + '...'
    for drug in drug_id:
        # get subframe and T it!
        current_res = df(grouped_res.loc[drug]).T

        # get responses data of cell lines in expr matrix
        cell_overlapped = []
        for cell in list(current_res.columns):
            if cell in list(expr.columns):
                cell_overlapped.append(cell)
            else:
                continue

        current_res = df(current_res.loc[:, cell_overlapped])
        current_expr = df(expr.loc[:, cell_overlapped])

        for pharm in sub_pharm:
            '''
            Parameter optimization.
            '''
            # print np.ravel(df(current_res.loc[pharm, :]).values)
            op_alpha, op_l1_ratio = paraEN(
                current_expr=current_expr,
                current_res=df(current_res.loc[pharm, :]),
                pharm=pharm, drug=drug)

            '''
            Bootstrapping for 200 times
            '''
            score_bs = bootPred(
                current_expr=current_expr,
                current_res=df(current_res.loc[pharm, :]),
                op_alpha=op_alpha,
                op_l1_ratio=op_l1_ratio)
            # score_bs = bootPred(current_expr=current_expr, current_res=df(current_res.loc[pharm, :]), op_alpha=0.2, op_l1_ratio=0.1)

            score_bs.to_csv(
                'Elastic_net_para/bootstrap_performance/' + pharm + '/' + str(drug) + '.csv',
                sep=',')

    return


def main():
    '''
    Main structure of CosEN.py would be built here.
    '''
    print 'CosEN beginning with data preprocessing...'

    '''
    Read in and pre-processing...
    '''
    # Read in expression matrix: 3333 rows(genes) x 505 columns(cell lines with cosmic id)
    expr = pd.read_csv('Expression/cancer_related_lncRNA_z_expr.csv', header=0, index_col=0, sep=',')

    # Read in drug response data: 224510 rows (pandas ID) x 4 columns (drug id, cell line id, LN_IC50, AUC)
    drug_res = pd.read_csv('drug_res/drug_res_merge.csv', header=0, sep='\t')
    # turn ID into str
    drug_res[['DRUG_ID', 'COSMIC_ID']] = drug_res[['DRUG_ID', 'COSMIC_ID']].astype(str)

    grouped_res = drug_res.copy()
    # Shape an hierarchical data frame
    grouped_res = grouped_res.groupby(['DRUG_ID', 'COSMIC_ID']).sum()

    """""""""""""""""""""""""""
    '''Elastic net regression'''
    """""""""""""""""""""""""""
    print """Usage: ./ CosEN_para.py [ncpus]
    [ncpus] - the number of workers to run in parallel, 
    if omitted it will be set to the number of processors in the system
    """

    # sub_pharm = ['LN_IC50', 'AUC']
    sub_pharm = ['LN_IC50']
    drug_id = list(set(drug_res.loc[:, 'DRUG_ID']))

    # tuple of all parallel python servers to connect with
    ppservers = ()
    # ppservers = ("10.0.0.1",)

    if len(sys.argv) > 1:
        ncpus = int(sys.argv[1])
        # Creates jobserver with ncpus workers
        job_server = pp.Server(ncpus, ppservers=ppservers)
    else:
        # Creates jobserver with automatically detected number of workers
        job_server = pp.Server(ppservers=ppservers)
    print "Starting Elastic net regression with", job_server.get_ncpus(), "workers"

    # divided the response into x blocker, where x equals specified number of cores
    # here we specify x as 5
    drug_id_0 = drug_id[:50]
    drug_id_1 = drug_id[50:100]
    drug_id_2 = drug_id[100:150]
    drug_id_3 = drug_id[150:200]
    drug_id_4 = drug_id[200:]

    # submit the jobs
    job_0 = job_server.submit(
        runEN,
        (grouped_res, drug_id_0, expr, sub_pharm, '0',),
        (paraEN, bootstrapResample, bootPred,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.linear_model import enet_path",
            "import matplotlib.pyplot as plt",
            "import matplotlib",))
    job_1 = job_server.submit(
        runEN,
        (grouped_res, drug_id_1, expr, sub_pharm, '1',),
        (paraEN, bootstrapResample, bootPred,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.linear_model import enet_path",
            "import matplotlib.pyplot as plt",
            "import matplotlib",))
    job_2 = job_server.submit(
        runEN,
        (grouped_res, drug_id_2, expr, sub_pharm, '2',),
        (paraEN, bootstrapResample, bootPred,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.linear_model import enet_path",
            "import matplotlib.pyplot as plt",
            "import matplotlib",))
    job_3 = job_server.submit(
        runEN,
        (grouped_res, drug_id_3, expr, sub_pharm, '3',),
        (paraEN, bootstrapResample, bootPred,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.linear_model import enet_path",
            "import matplotlib.pyplot as plt",
            "import matplotlib",))
    job_4 = job_server.submit(
        runEN,
        (grouped_res, drug_id_4, expr, sub_pharm, '4',),
        (paraEN, bootstrapResample, bootPred,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.linear_model import enet_path",
            "import matplotlib.pyplot as plt",
            "import matplotlib",))
    job_0()
    job_1()
    job_2()
    job_3()
    job_4()

    print 'Now the war is end...OvO!'


if __name__ == '__main__':
    main()
