#!/usr/bin/env python
'''
CosEN_para.py is analogus to CosEN.py but in a parallel running fashion.
Input:
    Expression matrix: Expression/filtered_lncRNA_z_EXP.csv
    Drug responses data: drug_res/drug_res_merge.csv
Operation: (*lead to one or multi output)
    1. Parameter Optimization*
        Output: optimized parameter for 265 drugs (IC50 and AUC)
    2. Bootstrapping
        Using optimized parameters, compute the predictive scores for each drug (IC50 and AUC)
    3. Get top predictive genes*
        Sort the genes by predictive scores and get top 20 genes.
        Output: predictive score of all genes for each drug (IC50 and AUC)
    4. Final model construction*:
        i. Get coefficients
            Output: Average coefficiencts of every predictors in each model (IC50 and AUC)
        ii. Model performance assessments
            Pearson's correlation coefficients, kendall's tau
'''

import pp
import sys
import numpy as np
import pandas as pd
import scipy.stats as stats

from pandas import DataFrame as df
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import ElasticNetCV
from sklearn.cross_validation import cross_val_predict
from sklearn.cross_validation import KFold
from sklearn.metrics import r2_score

__Author__ = 'Yue Wang'
__Date__ = 'Nov 23, 2016'


def buildEN(op_alpha, op_l1_ratio, gene_expr, drug_res):
    '''
    Build the model with 10-fold cv and optimized parameters.
    Input:
        op_alpha, op_l1_ratio: float, optimized parameters
        gene_expr: np.array, n row, m col, where m is the number of selected genes
        drug_res: np.array, n col, 0 row
    Return:
        coeff_matrix: list, coefficients of genes in each iteration
        performance: list, average performance by cv, Pearson's correlation r and Kendall's tau between predicted values and real values
    '''
    # print gene_expr.shape, drug_res.shape
    assert gene_expr.shape[0] == drug_res.shape[0]
    n_samples = gene_expr.shape[0]

    '''
    Coefficient training...
    '''
    # Initial output matrix
    coeff_matrix = []

    # Generate 10 fold selection
    kf = KFold(n_samples, n_folds=10)
    # Initial score list for performance by CV
    score_cv = []

    # Split training set and test set, and training for 10 interations in total
    print 'Spliting training set and test set, using 10-fold...'
    for train_index, test_index in kf:
        X_train, X_test = gene_expr[train_index], gene_expr[test_index]
        y_train, y_test = drug_res[train_index], drug_res[test_index]

        # solve elastic net equation
        enet = ElasticNet(alpha=op_alpha, l1_ratio=op_l1_ratio)
        enet.fit(X_train, y_train)

        # output the coefficients
        coeff_matrix.append(list(enet.coef_))

        # assess the performance of this iteration
        y_pred_enet = enet.predict(X_test)
        score_cv.append(r2_score(y_test, y_pred_enet))

    '''
    One way prediction by cross_val_predict
    '''
    enet_for_pred = ElasticNet(alpha=op_alpha, l1_ratio=op_l1_ratio)
    # 10 iterations of k-fold cv
    pred = []
    for i in range(10):
        cv = KFold(n_samples, n_folds=10, shuffle=True)
        pred.append(cross_val_predict(enet_for_pred, gene_expr, drug_res, cv=cv))

    print 'Assessing model performance...'

    # Pearson's correlation coefficient r
    r_list = []
    tau_list = []
    for p in pred:
        r, p_2_tail = stats.pearsonr(drug_res, p)
        r_list.append(r)

        # Kendall's tau
        tau, p_value = stats.kendalltau(list(drug_res), list(p))
        tau_list.append(tau)

    # Assign output
    cv_m = np.mean(score_cv)
    r_m = np.mean(r_list)
    t_m = np.mean(tau_list)
    performance = [score_cv, r_list, tau_list]
    performance_m = [cv_m, r_m, t_m]

    return coeff_matrix, performance, performance_m, pred


def bootPred(current_expr, current_res, op_alpha, op_l1_ratio):
    '''
    Bootstrapping and compute the predictive scores.
    Input:
        current_expr, current_res: pandas dataframe, row = gene/pharm, col = expression/response across cell lines
    Operation:
        1. Array reshape
        2. Bootstrap resample
        3. Solve elastic net regression equation for each resampled set
        4. Compute the predictive scores
    Output:
        pred_frame: pandas dataframe, row = genes, col = predictive scores
    '''
    # array reshape
    array_expr = np.array(current_expr.values.T)
    array_res = np.ravel(current_res.values)

    # initial the coefficent matrix
    beta_matrix = []

    for i in range(200):
        # resample
        expr, res = bootstrapResample(array_expr, array_res, n=None)

        # Training
        enet = ElasticNet(alpha=op_alpha, l1_ratio=op_l1_ratio)
        enet.fit(expr, res)

        beta_matrix.append(enet.coef_)

    print 'Computing predictive scores...'
    # compute the predictive score
    beta_array = predictiveIndicator(beta_matrix=beta_matrix)
    gene_id = list(current_expr.index)

    # format return frame
    pred_frame = df(beta_array, index=gene_id, columns=['predictive_score'])

    # sort by descending
    pred_frame = pred_frame.sort_values(by='predictive_score', ascending=False)

    return pred_frame


def predictiveIndicator(beta_matrix):
    '''
    Compute predictive scores for each gene:
        For each column(a gene's coefficients across all resampled solution):
            sum(0 if x = 0 else 1)
    beta_matrix: multidimensional list-like, dtype = float
    Return:
        beta_array: one-dimentional list-like, dtype = float
    '''
    beta_array = []
    tbeta_matrix = np.array(beta_matrix, dtype=float).transpose()

    # sum the indicator
    for gene in tbeta_matrix:
        s = 0
        for i in gene:
            # if being pulled to zero in the model
            if i == 0:
                s += 0
            # if not being pulled to zero
            else:
                s += 1

        # add on to output
        beta_array.append(float(s) / 200)

    return beta_array


def bootstrapResample(current_expr, current_res, n=None):
    '''
    A bootstrapping procedure to generate resampled datasets.
    current_expr, current_res: array-like matrix
    n: int, optional, length of resampled array, equal to len(data to resample) if n==None
    Results: resampled_expr, resampled_res
    '''
    # assign bootstrap resampled size
    if n is None:
        assert len(current_expr) == len(current_res)
        n = len(current_res)

    # generate resample index
    resample_i = np.floor(np.random.rand(n) * n).astype(int)
    resampled_expr = np.array([current_expr[i] for i in resample_i], dtype=float)
    resampled_res = [current_res[i] for i in resample_i]

    return resampled_expr, resampled_res


def paraEN(current_expr, current_res):
    '''
    Parameter optimization for elastic net regression model.
    ONLY require for models with genes selected by bootstrapping.
    Input:
        gene_expr: pandas dataframe, row = gene, col = gene expression matrix across all cell lines
        drug_res: pandas dataframe, row = pharm, col = drug response matrix across all cell lines
    Return:
        model.alpha_, model.l1_ratio_
    '''
    print 'Optimizing alpha and l1 ratio...'

    model = ElasticNetCV(l1_ratio=[.28, .36, .44, .52, .6, .68, .76, .84, .92, 1.0], n_alphas=200, max_iter=10000000, cv=10).fit(np.array(current_expr.values.T), np.ravel(current_res.values))
    # for test
    # model = ElasticNetCV(l1_ratio=[.28], n_alphas=1, max_iter=10000000, cv=5).fit(np.array(current_expr.values.T), np.ravel(current_res.values))

    # out put the parameters
    return model.alpha_, model.l1_ratio_


def runEN(grouped_res, drug_id, expr, sub_pharm, job_id):
    '''
    Skeletal of elastic net regression model: training, bootstrapping, construction
    Input:
        grouped_res: hierarchical pandas dataframe, index = drug, cell line, col = cell lines, value = IC50 or AUC
        drug_id: list, cosmic of drugs in this running
        expr: pandas dataframe, index = genes, col = cell lines, value = expression
        sub_pharm: list, 'LN_IC50' and 'AUC'
        job_id: str, parallel parameter indicating the id of processor used
    Output:
        final_model_performance: csv file, performance of all of the models in this running, given job_id
        predictive_scores: csv file
        final model_coefficients: csv file
    '''
    print 'Submitting to processor' + job_id + '...'
    final_model_performance = {}
    final_model_performance['LN_IC50'] = {}
    final_model_performance['AUC'] = {}
    for drug in drug_id:
        # get subframe and T it!
        current_res = df(grouped_res.loc[drug]).T

        # get responses data of cell lines in expr matrix
        cell_overlapped = []
        for cell in list(current_res.columns):
            if cell in list(expr.columns):
                cell_overlapped.append(cell)
            else:
                continue

        current_res = df(current_res.loc[:, cell_overlapped])
        current_expr = df(expr.loc[:, cell_overlapped])

        for pharm in sub_pharm:
            '''
            Parameter optimization.
            '''
            # print np.ravel(df(current_res.loc[pharm, :]).values)
            op_alpha, op_l1_ratio = paraEN(current_expr=current_expr, current_res=df(current_res.loc[pharm, :]))

            '''
            Bootstrapping for 200 times to find out top predictive genes
            '''
            pred_gene = bootPred(current_expr=current_expr, current_res=df(current_res.loc[pharm, :]), op_alpha=op_alpha, op_l1_ratio=op_l1_ratio)
            pred_gene.to_csv('Elastic_net_para/predictive_scores/' + pharm + '/' + drug + '.csv', sep='\t')

            '''
            Get top 20 predictive genes.
            '''
            top_expr = df(current_expr.loc[list(pred_gene.index)[:20], :])

            '''
            Build final model and get coefficients.
            '''
            print 'Building final model of ' + drug + ' ' + pharm + '...'
            # Parameter optimization
            op_alpha, op_l1_ratio = paraEN(current_expr=top_expr, current_res=df(current_res.loc[pharm, :]))

            # Get coefficients
            coeff_matrix, performance, performance_m, pred = buildEN(op_alpha=op_alpha, op_l1_ratio=op_l1_ratio, gene_expr=top_expr.values.T, drug_res=np.ravel(df(current_res.loc[pharm, :]).values))

            '''
            Output the predicted values.
            '''
            o = open('Elastic_net_para/predicted_values/' + pharm + '_' + drug + '.csv', 'w')
            cell_line = list(current_res.columns)
            o.write('\t'.join(cell_line) + '\n')
            for p in pred:
                o.write('\t'.join([str(p[i]) for i in range(len(p))]) + '\n')
            o.close()

            '''
            Output the performance
            '''
            final_model_performance[pharm][drug] = performance_m

            o = open('Elastic_net_para/final_model_performance/' + pharm + '_' + drug + '.csv', 'w')
            o.write('Average by CV' + '\t' + 'Pearson r' + '\t' + 'Kendall tau' + '\n')
            for j in range(len(performance[0])):
                o.write('\t'.join([str(performance[i][j]) for i in range(len(performance))]) + '\n')
            o.close()

            '''
            Output the coefficients
            '''
            # Make output as pandas dataframe
            coef_frame = df(coeff_matrix, index=['1', '2', '3', '4', '5', '6', '7', '8', '9', '10'], columns=list(top_expr.index))
            # get average value of coefficients
            coef_mean = df(coef_frame.mean(), columns=['average_coefficient'])
            # sort by coefficients: ascending
            coef_mean = coef_mean.sort_values(by='average_coefficient')
            # output
            coef_mean.to_csv('Elastic_net_para/final_model_coefficients/' + pharm + '/' + drug + '.csv', sep='\t')

    '''
    Output the final model performance for each block.
    '''
    print 'Output the final model performance...'

    for pharm in sub_pharm:
        per = df(final_model_performance[pharm].values(), index=final_model_performance[pharm].keys(), columns=['Average by CV', 'Pearson r', 'Kendall tau'])
        per.to_csv('Elastic_net_para/final_model_performance/' + pharm + '_block_' + job_id + '.csv', sep='\t')


def main():
    '''
    Main structure of CosEN.py would be built here.
    '''
    print 'CosEN beginning with data preprocessing...'

    '''
    Read in and pre-processing...
    '''
    # Read in expression matrix: 3333 rows(genes) x 505 columns(cell lines with cosmic id)
    expr = pd.read_csv('Expression/cancer_related_lncRNA_z_expr.csv', header=0, index_col=0, sep=',')
    print expr.head()

    # Read in drug response data: 224510 rows (pandas ID) x 4 columns (drug id, cell line id, LN_IC50, AUC)
    drug_res = pd.read_csv('drug_res/drug_res_merge.csv', header=0, sep='\t')
    # turn ID into str
    drug_res[['DRUG_ID', 'COSMIC_ID']] = drug_res[['DRUG_ID', 'COSMIC_ID']].astype(str)

    grouped_res = drug_res.copy()
    # Shape an hierarchical data frame
    grouped_res = grouped_res.groupby(['DRUG_ID', 'COSMIC_ID']).sum()

    """""""""""""""""""""""""""
    '''Elastic net regression'''
    """""""""""""""""""""""""""
    print """Usage: ./ CosEN_para.py [ncpus]
    [ncpus] - the number of workers to run in parallel, 
    if omitted it will be set to the number of processors in the system
    """

    sub_pharm = ['LN_IC50', 'AUC']
    drug_id = list(set(drug_res.loc[:, 'DRUG_ID']))

    # tuple of all parallel python servers to connect with
    ppservers = ()
    # ppservers = ("10.0.0.1",)

    if len(sys.argv) > 1:
        ncpus = int(sys.argv[1])
        # Creates jobserver with ncpus workers
        job_server = pp.Server(ncpus, ppservers=ppservers)
    else:
        # Creates jobserver with automatically detected number of workers
        job_server = pp.Server(ppservers=ppservers)
    print "Starting Elastic net regression with", job_server.get_ncpus(), "workers"

    # divided the response into x blocker, where x equals specified number of cores
    # here we specify x as 5
    drug_id_0 = drug_id[:50]
    drug_id_1 = drug_id[50:100]
    drug_id_2 = drug_id[100:150]
    drug_id_3 = drug_id[150:200]
    drug_id_4 = drug_id[200:]

    # submit the jobs
    job_0 = job_server.submit(
        runEN,
        (grouped_res, drug_id_0, expr, sub_pharm, '0',),
        (paraEN, bootstrapResample, predictiveIndicator, bootPred, buildEN,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.cross_validation import cross_val_predict",
            "from sklearn.cross_validation import KFold",
            "from sklearn.metrics import r2_score",))
    job_1 = job_server.submit(
        runEN,
        (grouped_res, drug_id_1, expr, sub_pharm, '1',),
        (paraEN, bootstrapResample, predictiveIndicator, bootPred, buildEN,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.cross_validation import cross_val_predict",
            "from sklearn.cross_validation import KFold",
            "from sklearn.metrics import r2_score",))
    job_2 = job_server.submit(
        runEN,
        (grouped_res, drug_id_2, expr, sub_pharm, '2',),
        (paraEN, bootstrapResample, predictiveIndicator, bootPred, buildEN,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.cross_validation import cross_val_predict",
            "from sklearn.cross_validation import KFold",
            "from sklearn.metrics import r2_score",))
    job_3 = job_server.submit(
        runEN,
        (grouped_res, drug_id_3, expr, sub_pharm, '3',),
        (paraEN, bootstrapResample, predictiveIndicator, bootPred, buildEN,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.cross_validation import cross_val_predict",
            "from sklearn.cross_validation import KFold",
            "from sklearn.metrics import r2_score",))
    job_4 = job_server.submit(
        runEN,
        (grouped_res, drug_id_4, expr, sub_pharm, '4',),
        (paraEN, bootstrapResample, predictiveIndicator, bootPred, buildEN,),
        (
            "import numpy as np",
            "import pandas as pd",
            "import scipy.stats as stats",
            "from pandas import DataFrame as df",
            "from sklearn.linear_model import ElasticNet",
            "from sklearn.linear_model import ElasticNetCV",
            "from sklearn.cross_validation import cross_val_predict",
            "from sklearn.cross_validation import KFold",
            "from sklearn.metrics import r2_score",))
    job_0()
    job_1()
    job_2()
    job_3()
    job_4()

    print 'Now the war is end...OvO!'


if __name__ == '__main__':
    main()
